﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ModelExchanger.AnalysisDataModel;
using ModelExchanger.AnalysisDataModel.Loads;    //gv added
using ModelExchanger.AnalysisDataModel.Models;
using ModelExchanger.AnalysisDataModel.StructuralElements;
using SCIA.OpenAPI;
using SCIA.OpenAPI.Results;  //gv added
using Results64Enums;  //gv added
using System.IO;

namespace SciaEngineer.OpenApi.SEN22_0
{

    public enum ResultsFor { Elem_0D, Elem_1D, Elem_2D }
    public enum ResultsType { InternalForces, Deformations }
    public enum ResultsCase { Loadcase, Combination }
    public enum OutputContent { Default, Extended, Extended_v2 }
    public enum OutputLocation { Screen, Txt_file }


    public static class gvResults
    {

        public static void OutputResults(EsaProject proj, AnalysisModel model,
                                         OutputContent outputContent, OutputLocation outputLocation,
                                         ResultsFor resultsFor, ResultsType resultsType, ResultsCase resultsCase)
        {
            ResultsAPI rapi = proj.Model.InitializeResultsAPI();
            if (rapi is null)
            {
                return;
            }


            ResultKey ResKey = new ResultKey();

            IReadOnlyList<IAnalysisObject> Elems;
            switch (resultsFor)
            {
                case ResultsFor.Elem_0D:
                    Elems = model.OfType<StructuralPointSupport>().ToList();
                    ResKey.EntityType = eDsElementType.eDsElementType_PointSupportPoint;
                    ResKey.Dimension = eDimension.eDim_reactionsPoint;
                    ResKey.ResultType = eResultType.eReactionsNodes;
                    ResKey.CoordSystem = eCoordSystem.eCoordSys_Global;
                    break;
                default:
                case ResultsFor.Elem_1D:
                    Elems = model.OfType<StructuralCurveMember>().ToList();
                    ResKey.EntityType = eDsElementType.eDsElementType_Beam;
                    ResKey.Dimension = eDimension.eDim_1D;
                    switch (resultsType)
                    {
                        default:
                        case ResultsType.InternalForces:
                            ResKey.ResultType = eResultType.eFemBeamInnerForces;
                            ResKey.CoordSystem = eCoordSystem.eCoordSys_Local;
                            break;
                        case ResultsType.Deformations:
                            ResKey.ResultType = eResultType.eFemBeamDeformation;
                            ResKey.CoordSystem = eCoordSystem.eCoordSys_Local;
                            break;
                    }
                    break;
                case ResultsFor.Elem_2D:
                    Elems = model.OfType<StructuralSurfaceMember>().ToList();
                    ResKey.EntityType = eDsElementType.eDsElementType_Slab;
                    ResKey.Dimension = eDimension.eDim_2D;
                    switch (resultsType)
                    {
                        default:
                        case ResultsType.InternalForces:
                            ResKey.ResultType = eResultType.eFemInnerForces;
                            ResKey.CoordSystem = eCoordSystem.eCoordSys_Local;
                            break;
                        case ResultsType.Deformations:
                            ResKey.ResultType = eResultType.eFemDeformations;
                            ResKey.CoordSystem = eCoordSystem.eCoordSys_Local;
                            break;
                    }
                    break;
            }

            IReadOnlyList<IAnalysisObject> Cases;
            switch (resultsCase)
            {
                default:
                case ResultsCase.Loadcase:
                    ResKey.CaseType = eDsElementType.eDsElementType_LoadCase;
                    Cases = model.OfType<StructuralLoadCase>().ToList();
                    break;
                case ResultsCase.Combination:
                    ResKey.CaseType = eDsElementType.eDsElementType_Combination;
                    Cases = model.OfType<StructuralLoadCombination>().ToList();
                    break;
            }




            // loop over all cases and all elems
            foreach (var load in Cases)
            {
                ResKey.CaseId = load.Id;
                foreach (var elem in Elems)
                {
                    ResKey.EntityName = elem.Name;
                    OutputWriter(ResKey, load.Name, rapi, outputContent, outputLocation);
                    
                    //ResultBlock resblock = new ResultBlock();
                    //resblock = ReadResults(ResKey, rapi, load.Name);
                    //ResultBlocks.Add(resblock);

                    ResultBlocks.Add(ReadResults(ResKey, rapi, load.Name));
                }
            }
        }

        public static List<ResultBlock> ResultBlocks = new List<ResultBlock>();


        private static void OutputWriter(ResultKey ResKey, string loadname, ResultsAPI rapi, OutputContent outputContent, OutputLocation outputLocation)
        {
            StringBuilder sb = new StringBuilder();

            Result Res;
            Res = rapi.LoadResult(ResKey);

            switch (outputContent)
            {
                default:
                case OutputContent.Default:
                    sb.AppendLine();
                    sb.AppendLine("======================");
                    sb.AppendLine($"---  Results for case < {loadname} > - elem < {ResKey.EntityName} >");
                    sb.AppendLine("======================");
                    sb.AppendLine(Res.GetTextOutput());
                    break;
                case OutputContent.Extended:
                    sb.AppendLine();
                    sb.AppendLine("======================");
                    sb.AppendLine($"---  Results for case < {loadname} > - elem < {ResKey.EntityName} >");
                    sb.AppendLine("======================");
                    GetExtendedTextOutput(ref sb, Res, rapi, ResKey);
                    break;
            }

            switch (outputLocation)
            {
                default:
                case OutputLocation.Screen:
                    Console.WriteLine(sb.ToString());
                    break;
                case OutputLocation.Txt_file:
                    using (StreamWriter sw = new StreamWriter(Path.Combine(Program.SenTempFolder, "ExtendedOutput.txt"), append: true))
                    {
                        sw.WriteLine(sb.ToString());
                    }
                    Console.WriteLine(sb.ToString());
                    break;
            }
        }


        private static ResultBlock ReadResults(ResultKey ResKey, ResultsAPI rapi, string loadname)
        {
            int sectionCount = 0;
            int magnitudesCount = 0;
            Result Res;
            Dictionary<int, string> Pos = new Dictionary<int, string>();

            // get the results
            Res = rapi.LoadResult(ResKey);
            // get the layersCount
            int layersCount = Res.GetLayersCount();

            switch (ResKey.ResultType)
            {
                // 0D results
                case eResultType.eReactionsNodes:
                    magnitudesCount = Res.GetMagnitudesCount();
                    sectionCount = 1;
                    Pos.Add(0, "(N/A)");
                    break;
                // 1D results
                case eResultType.eFemBeamDeformation:
                case eResultType.eFemBeamInnerForces:
                    magnitudesCount = Res.GetMagnitudesCount() - 1;  // don't need the last magnitude 'Vr'
                    sectionCount = Res.GetSectionCount();
                    for (int i = 0; i < sectionCount; i++)
                    {
                        Pos.Add(i, string.Format("({0:0.000})", Res.GetSectionPosX(i)));
                    }
                    break;
                // 2D  Results
                case eResultType.eFemInnerForces:
                case eResultType.eFemDeformations:
                    magnitudesCount = Res.GetMagnitudesCount();
                    sectionCount = Res.GetMeshElementCount();
                    for (int i = 0; i < sectionCount; i++)
                    {
                        Pos.Add(i, string.Format("({0:0.000},{1:0.000},{2:0.000})", rapi.GetMeshElementPosition(Res, i).X, rapi.GetMeshElementPosition(Res, i).Y, rapi.GetMeshElementPosition(Res, i).Z));
                    }
                    break;
            }


            // header
            string[] ValsHeaders = new string[magnitudesCount];
            for (int magnitude = 0; magnitude < magnitudesCount; magnitude++)
            {
                ValsHeaders[magnitude] = Res.GetMagnitudeName(magnitude);
            }
            ResultHeader resultHeader = new ResultHeader(_ElemHeader: "Elem", _CaseHeader: "Case", _MinMaxHeader: "MinMax", _PosHeader: "Pos", _ValsHeaders: ValsHeaders);


            List<ResultLine> ResultValues = new List<ResultLine>();
            string minmax = "";
            for (int layer = 0; layer < layersCount; layer++)
            {
                // combination layer
                if (layersCount > 1)
                {
                    minmax = (layer % 2) == 0 ? $"Max {Res.GetMagnitudeName(layer / 2)}" : $"Min {Res.GetMagnitudeName(layer / 2)}";
                }

                // values
                for (int section = 0; section < sectionCount; section++)
                {
                    double[] Vals = new double[magnitudesCount];
                    for (int magnitude = 0; magnitude < magnitudesCount; magnitude++)
                    {
                        Vals[magnitude] = Res.GetValue(magnitude, section);
                    }
                    ResultValues.Add(new ResultLine(_ElemName: ResKey.EntityName, _CaseName: loadname, _MinMax: minmax, _Pos: Pos[section], _Vals: Vals));
                }
            }

            return new ResultBlock(_resultHeader: resultHeader, _resultLines: ResultValues);
        }


        private static void GetExtendedTextOutput(ref StringBuilder sb, Result Res, ResultsAPI rapi, ResultKey Res_Key)
        {
            double Val = 0.0;

            int sectionCount = 0;
            int magnitudesCount = 0;
            int layersCount = 0;

            string fmtElemStr = "";
            string fmtElemVal = "";
            string fmtDblStr = "";
            string fmtDblVal = "";
            double scaleFactor = 1.0;

            //ResultKey Res_Key = Res.GetResultKey();

            // get the magnitudesCount
            magnitudesCount = Res.GetMagnitudesCount();
            // get the layersCount
            layersCount = Res.GetLayersCount();



            // defining the sectionCount
            Dictionary<int, string> XYZ = new Dictionary<int, string>();

            Dictionary<int, string> Pos = new Dictionary<int, string>();

            switch (Res_Key.ResultType)
            {
                // 0D results
                case eResultType.eReactionsNodes:
                    magnitudesCount = Res.GetMagnitudesCount();
                    sectionCount = 1;
                    XYZ.Add(0, "   (N/A)");
                    Pos.Add(0, "(N/A)");
                    break;
                // 1D results
                case eResultType.eFemBeamDeformation:
                case eResultType.eFemBeamInnerForces:
                    magnitudesCount = Res.GetMagnitudesCount() - 1;  // don't need the last magnitude 'Vr'
                    sectionCount = Res.GetSectionCount();
                    for (int i = 0; i < sectionCount; i++)
                    {
                        XYZ.Add(i, string.Format("   ({0:0.000})", Res.GetSectionPosX(i)));
                        Pos.Add(i, string.Format("({0:0.000})", Res.GetSectionPosX(i)));
                    }
                    break;
                // 2D  Results
                case eResultType.eFemInnerForces:
                case eResultType.eFemDeformations:
                    magnitudesCount = Res.GetMagnitudesCount();
                    sectionCount = Res.GetMeshElementCount();
                    for (int i = 0; i < sectionCount; i++)
                    {
                        XYZ.Add(i, string.Format("   ({0:0.000},{1:0.000},{2:0.000})", rapi.GetMeshElementPosition(Res, i).X, rapi.GetMeshElementPosition(Res, i).Y, rapi.GetMeshElementPosition(Res, i).Z));
                        Pos.Add(i, string.Format("({0:0.000},{1:0.000},{2:0.000})", rapi.GetMeshElementPosition(Res, i).X, rapi.GetMeshElementPosition(Res, i).Y, rapi.GetMeshElementPosition(Res, i).Z));
                    }
                    break;
            }

            // defining the output format
            fmtElemStr = "{0,6}";
            fmtElemVal = "{0,6:0}";
            fmtDblStr = "{0,14}";
            switch (Res_Key.ResultType)
            {
                // Internal Forces
                case eResultType.eReactionsNodes:
                case eResultType.eFemBeamInnerForces:
                case eResultType.eFemInnerForces:
                    fmtDblVal = "{0,14:0.00}";
                    scaleFactor = 0.0001;   // Newton -> kiloNewton
                    break;
                // Deformations
                case eResultType.eFemBeamDeformation:
                case eResultType.eFemDeformations:
                    fmtDblVal = "{0,14:0.000}";
                    scaleFactor = 1000.0;  // meter -> millimeter
                    break;
                // the Exceptions
                default:
                    fmtDblVal = "{0}";
                    scaleFactor = 1.0;
                    break;
            }


            // printing the formatted output tables
            using (StringWriter sw = new StringWriter())
            {
                // minmax layer
                string minmax = "";
                for (int layer = 0; layer < layersCount; layer++)
                {
                    // 

                    // combination layer
                    if (layersCount > 1)
                    {
                        minmax = (layer % 2) == 0 ? $"Max {Res.GetMagnitudeName(layer / 2)}" : $"Min {Res.GetMagnitudeName(layer / 2)}";
                        sw.WriteLine($"layer {minmax}");
                    }

                    // header
                    sw.Write(fmtElemStr, "Elem");
                    for (int magnitude = 0; magnitude < magnitudesCount; magnitude++)
                    {
                        sw.Write(fmtDblStr, Res.GetMagnitudeName(magnitude));
                    }
                    sw.Write("   Location");
                    sw.WriteLine();


                    // values
                    for (int section = 0; section < sectionCount; section++)
                    {
                        double[] Vals = new double[magnitudesCount];
                        sw.Write(fmtElemVal, section);
                        for (int magnitude = 0; magnitude < magnitudesCount; magnitude++)
                        {
                            Val = Res.GetValue(magnitude, section);
                            sw.Write((Val > 9.0e20) ? string.Format(fmtDblStr, "N/A") : string.Format(fmtDblVal, Val * scaleFactor));
                            Vals[magnitude] = Val * scaleFactor;
                        }
                        sw.Write(XYZ[section]);
                        sw.WriteLine();

                        //ResultValues.Add(new ResultLine(_ElemName: Res_Key.EntityName, _CaseName: "casename", _MinMax: minmax, _Pos: Pos[section], _Vals: Vals ));

                    }
                }
                sb.Append(sw.ToString());
            }
        }




































   
    }


    // Elem 1D
    // elem - case/combi - minmax - pos - Nx Vy Vz Mx My Mz
    // elem - case/combi - minmax - pos - Ux Uy Uz Fix Fiy Fiz
    // Elem 2D
    // elem - case/combi - minmax - posX posY posZ - mx my mxy vx vy nx ny nxy
    // elem - case/combi - minmax - posX posY posZ - Ux Uy Uz Fix Fiy Fiz


    public struct ResultLine
    {
        public ResultLine(string _ElemName, string _CaseName, string _MinMax, string _Pos, double [] _Vals)
        {
            ElemName = _ElemName; CaseName = _CaseName; MinMax = _MinMax; Pos = _Pos; Vals = _Vals;
        }
        public string ElemName { get; }
        public string CaseName { get; }
        public string MinMax { get; }
        public string Pos { get; }
        public double [] Vals { get; }
    }
    public struct ResultHeader
    {
        public ResultHeader(string _ElemHeader, string _CaseHeader, string _MinMaxHeader, string _PosHeader, string [] _ValsHeaders)
        {
            ElemHeader = _ElemHeader; CaseHeader = _CaseHeader; MinMaxHeader = _MinMaxHeader; PosHeader = _PosHeader; ValsHeaders = _ValsHeaders; 
        }
        public string ElemHeader { get; }
        public string CaseHeader { get; }
        public string MinMaxHeader { get; }
        public string PosHeader { get; }
        public string [] ValsHeaders { get; }
    }
    public struct ResultBlock
    {
        public ResultBlock(ResultHeader _resultHeader, List<ResultLine> _resultLines)
        {
            resultHeader = _resultHeader;  resultLines = _resultLines;
        }
        public ResultHeader resultHeader { get; }
        public List<ResultLine> resultLines { get; }
    }





}


